// dollarcent_test.cpp
// Glenn G. Chappell
// 3 Oct 2013
//
// For CS 202 Fall 2013
// Test program for class DollarCent
// Used in Assignment 4, Exercise A

// Includes for code to be tested
#include "dollarcent.h" // For class DollarCent
#include "dollarcent.h" // Double inclusion test

// Includes for testing package & code common to all test programs
#include <iostream>     // for std::cout, std::endl, std::cin
#include <string>       // for std::string
#include <stdexcept>    // for std::runtime_error

// Additional includes for this test program
#include <sstream>      // for std::ostringstream


// ************************************************************************
// Testing Package:
//     Class Tester - For Tracking Tests
// ************************************************************************


// class Tester
// For extremely simple unit testing.
// Keeps track of number of tests and number of passes.
// Use test (with success/failure parameter) to do a test.
// Get results with numTests, numPassed, numFailed, allPassed.
// Restart testing with reset.
// Invariants:
//     countTests_ == number of tests (calls to test) since last reset.
//     countPasses_ == number of times function test called with true param
//      since last reset.
//     0 <= countPasses_ <= countTests_.
//     tolerance_ >= 0.
class Tester {

// ***** Tester: ctors, dctor, op= *****
public:

    // Default ctor
    // Sets countTests_, countPasses_ to zero, tolerance_ to given value
    // Pre: None.
    // Post:
    //     numTests == 0, countPasses == 0, tolerance_ == abs(theTolerance)
    // Does not throw (No-Throw Guarantee)
    Tester(double theTolerance = 0.0000001)
        :countTests_(0),
         countPasses_(0),
         tolerance_(theTolerance >= 0 ? theTolerance : -theTolerance)
    {}

    // Compiler-generated copy ctor, copy op=, dctor are used

// ***** Tester: general public functions *****
public:

    // test
    // Handles single test, param indicates pass/fail
    // Pre: None.
    // Post:
    //     countTests_ incremented
    //     countPasses_ incremented if (success)
    //     Message indicating test name (if given)
    //      and pass/fail printed to cout
    // Does not throw (No-Throw Guarantee)
    //  - Assuming exceptions have not been turned on for cout.
    void test(bool success,
              const std::string & testName = "")
    {
        ++countTests_;
        if (success) ++countPasses_;

        std::cout << "    ";
        if (testName != "")
        {
            std::cout << "Test: "
                      << testName
                      << " - ";
        }
        std::cout << (success ? "passed" : "********** FAILED **********")
                  << std::endl;
    }

    // ftest
    // Does single floating-point test.
    // Tests passes iff difference of first two values is <= tolerance.
    // Pre: None.
    // Post:
    //     countTests_ incremented
    //     countPasses_ incremented if (abs(val1-val2) <= tolerance_)
    //     Message indicating test name (if given)
    //      and pass/fail printed to cout
    // Does not throw (No-Throw Guarantee)
    void ftest(double val1,
               double val2,
               const std::string & testName = "")
    { test(val1-val2 <= tolerance_ && val2-val1 <= tolerance_, testName); }

    // reset
    // Resets *this to default constructed state
    // Pre: None.
    // Post:
    //     countTests_ == 0, countPasses_ == 0
    // Does not throw (No-Throw Guarantee)
    void reset()
    {
        countTests_ = 0;
        countPasses_ = 0;
    }

    // numTests
    // Returns the number of tests that have been done since last reset 
    // Pre: None.
    // Post:
    //     return == countTests_
    // Does not throw (No-Throw Guarantee)
    int numTests() const
    { return countTests_; }

    // numPassed
    // Returns the number of tests that have passed since last reset
    // Pre: None.
    // Post:
    //     return == countPasses_
    // Does not throw (No-Throw Guarantee)
    int numPassed() const
    { return countPasses_; }

    // numFailed
    // Returns the number of tests that have not passed since last reset
    // Pre: None.
    // Post:
    //     return + countPasses_ == numTests_
    // Does not throw (No-Throw Guarantee)
    int numFailed() const
    { return countTests_ - countPasses_; }

    // allPassed
    // Returns true if all tests since last reset have passed
    // Pre: None.
    // Post:
    //     return == (countPasses_ == countTests_)
    // Does not throw (No-Throw Guarantee)
    bool allPassed() const
    { return countPasses_ == countTests_; }

    // setTolerance
    // Sets tolerance_ to given value
    // Pre: None.
    // Post:
    //     tolerance_ = abs(theTolerance)
    // Does not throw (No-Throw Guarantee)
    void setTolerance(double theTolerance)
    { tolerance_ = (theTolerance >= 0 ? theTolerance : -theTolerance); }

// ***** Tester: data members *****
private:

    int countTests_;    // Number of tests done since last reset
    int countPasses_;   // Number of tests passed since last reset
    double tolerance_;  // Tolerance for floating-point near-equality tests

};  // end class Tester


// ************************************************************************
// Testing Package:
//     Class TypeCheck - Helper Class for Type Checking
// ************************************************************************


// class TypeCheck
// This class exists in order to have static member function check, which
// takes a parameter of a given type, by reference. Objects of type
// TypeCheck<T> cannot be created.
// Usage:
//     TypeCheck<MyType>::check(x)
//     returns true if the type of x is (MyType) or (const MyType),
//     otherwise false.
// Invariants: None.
// Requirements on Types: None.
template<typename T>
class TypeCheck {

private:

    // Uncopyable class. Do not define copy ctor, copy assn.
    TypeCheck(const TypeCheck &);
    TypeCheck<T> & operator=(const TypeCheck &);

    // Compiler-generated dctor is used (but irrelevant).

public:

    // check
    // The function and function template below simulate a single function
    // that takes a single parameter, and returns true iff the parameter has
    // type T or (const T).

    // check (reference-to-const T)
    // Pre: None.
    // Post:
    //     Return is true.
    // Does not throw (No-Throw Guarantee)
    static bool check(const T & param)
    { return true; }

    // check (reference-to-const non-T)
    // Pre: None.
    // Post:
    //     Return is false.
    // Requirements on types: None.
    // Does not throw (No-Throw Guarantee)
    template <typename OtherType>
    static bool check(const OtherType & param)
    { return false; }

};  // End class TypeCheck


// ************************************************************************
// Testing Package:
//     Class Counter - Helper Class for Counting Calls & Objects, Throwing
// ************************************************************************


// class Counter
// Item type for counting ctor, dctor, op= calls, counting existing
//  objects, and possibly throwing on copy. Has operator< (which always
//  returns false), allowing it to be the value type of a sorted container.
// If static member copyThrow_ is set, then copy ctor and copy assn throw
//  std::runtime_error. Exception object's "what" member is set to "C" by
//  the copy ctor and "A" by copy assn.
// Increments static data member ctorCount_ on default construction and
//  successful copy construction. Increments static data member assnCount_
//  on successful copy assignment. Increments static data member
//  dctorCount_ on destruction.
// Increments static data member existing_ on construction, and decrements
//  it on destruction.
// Static data member maxExisting_ is highest value of existing_ since last
//  reset, or start of program if reset has never been called.
// Invariants:
//     Counter::existing_ is number of existing objects of this class.
//     Counter::ctorCount_ is number of successful ctor calls since
//      most recent call to reset, or start of program if reset has never
//      been called.
//     Counter::dctorCount_ is (similarly) number of dctor calls.
//     Counter::assnCount_ is (similarly) number of copy assn calls.
//     Counter::maxExisting_ is (similarly) highest value existing_ has
//      assumed.
class Counter {

// ***** Counter: Ctors, dctor, op= *****
public:

    // Default ctor
    // Pre: None.
    // Post:
    //     (ctorCount_ has been incremented.)
    //     (existing_ has been incremented.)
    // Does not throw (No-Throw Guarantee)
    Counter()
    {
        ++existing_;
        if (existing_ > maxExisting_)
            maxExisting_ = existing_;
        ++ctorCount_;
    }

    // Copy ctor
    // Throws std::runtime_error if copyThrow_.
    // Pre: None.
    // Post:
    //     (ctorCount_ has been incremented.)
    //     (existing_ has been incremented.)
    // May throw std::runtime_error
    // Strong Guarantee
    Counter(const Counter & other)
    {
        if (copyThrow_)
            throw std::runtime_error("C");
        ++existing_;
        if (existing_ > maxExisting_)
            maxExisting_ = existing_;
        ++ctorCount_;
    }

    // Copy assignment
    // Throws std::runtime_error if copyThrow_.
    // Pre: None.
    // Post:
    //     Return value is *this.
    //     (assnCount_ has been incremented.)
    // May throw std::runtime_error
    // Strong Guarantee
    Counter & operator=(const Counter & rhs)
    {
        if (copyThrow_)
            throw std::runtime_error("A");
        ++assnCount_;
        return *this;
    }

    // Dctor
    // Pre: None.
    // Post:
    //     (dctorCount_ has been incremented.)
    //     (existing_ has been decremented.)
    // Does not throw (No-Throw Guarantee)
    ~Counter()
    {
        --existing_;
        ++dctorCount_;
    }

// ***** Counter: Functions dealing with count *****
public:

    // reset
    // Pre: None.
    // Post:
    //     maxExisting_ == existing_.
    //     ctorCount_ == 0.
    //     dctorCount_ == 0.
    //     assnCount_ == 0.
    //     copyThrow_ == shouldThrow.
    // Does not throw (No-Throw Guarantee)
    static void reset(bool shouldThrow = false)
    {
        maxExisting_ = existing_;
        ctorCount_ = 0;
        dctorCount_ = 0;
        assnCount_ = 0;
        copyThrow_ = shouldThrow;
    }

    // getExisting
    // Pre: None.
    // Post:
    //     return == existing_.
    // Does not throw (No-Throw Guarantee)
    static int getExisting()
    { return existing_; }

    // getMaxExisting
    // Pre: None.
    // Post:
    //     return == maxExisting_.
    // Does not throw (No-Throw Guarantee)
    static int getMaxExisting()
    { return maxExisting_; }

    // getCtorCount
    // Pre: None.
    // Post:
    //     return == ctorCount_.
    // Does not throw (No-Throw Guarantee)
    static int getCtorCount()
    { return ctorCount_; }

    // getDctorCount
    // Pre: None.
    // Post:
    //     return == dctorCount_.
    // Does not throw (No-Throw Guarantee)
    static int getDctorCount()
    { return dctorCount_; }

    // getAssnCount
    // Pre: None.
    // Post:
    //     return == assnCount_.
    // Does not throw (No-Throw Guarantee)
    static int getAssnCount()
    { return assnCount_; }

    // setCopyThrow
    // Pre: None.
    // Post:
    //     copyThrow_ == shouldThrow
    // Does not throw (No-Throw Guarantee)
    static void setCopyThrow(bool shouldThrow)
    { copyThrow_ = shouldThrow; }

// ***** Counter: Data Members *****
private:

    static int existing_;     // # of existing objects
    static int maxExisting_;  // Max # of existing objects
    static int ctorCount_;    // # of successful (non-throwing) ctor calls
    static int dctorCount_;   // # of dctor calls
    static int assnCount_;    // # of successful (non-throwing) copy = calls
    static bool copyThrow_;   // true if copy operations (ctor, =) throw

};  // End class Counter

// Definition of static data member of class Counter
int Counter::existing_ = 0;
int Counter::maxExisting_ = 0;
int Counter::ctorCount_ = 0;
int Counter::dctorCount_ = 0;
int Counter::assnCount_ = 0;
bool Counter::copyThrow_ = false;


// operator< (Counter)
// Dummy-ish operator<, forming a strict weak order for Counter class
// Returns false (which is legal for a strict weak order; all objects of
//  type Counter are equivalent).
// Pre: None.
// Post:
//     Return value == false.
// Does not throw (No-Throw Guarantee)
bool operator<(const Counter & a,
               const Counter & b)
{ return false; }


// ************************************************************************
// Definitions for this test suite
// ************************************************************************


// recurse
// Given n, makes n recursive calls
void recurse(int n)
{
    if (n == 0)
        return;
    recurse(n-1);
}


// ************************************************************************
// Test Suite Functions
// ************************************************************************


// test_DollarCent_output
// Test suite for class DollarCent: output
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent_output(Tester & t)
{
    std::cout << "Test Suite: class DollarCent - output"
              << std::endl;

    {
    const DollarCent d;
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$0.00", "print $0.00 - #1");
    }

    {
    const DollarCent d(0);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$0.00", "print $0.00 - #2");
    }

    {
    const DollarCent d(0,0);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$0.00", "print $0.00 - #3");
    }

    {
    const DollarCent d(99);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$99.00", "print $99.00 - #1");
    }

    {
    const DollarCent d(99,0);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$99.00", "print $99.00 - #2");
    }

    {
    const DollarCent d(-99);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-99.00", "print $-99.00 - #1");
    }

    {
    const DollarCent d(-99,0);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-99.00", "print $-99.00 - #2");
    }

    {
    const DollarCent d(99,4);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$99.04", "print $99.04");
    }

    {
    const DollarCent d(-99,-4);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-99.04", "print $-99.04");
    }

    {
    const DollarCent d(99,-91);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$98.09", "print $98.09");
    }

    {
    const DollarCent d(-99,91);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-98.09", "print $-98.09");
    }

    {
    const DollarCent d(99,12);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$99.12", "print $99.12");
    }

    {
    const DollarCent d(-99,-12);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-99.12", "print $-99.12");
    }

    {
    const DollarCent d(99,-12);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$98.88", "print $98.88");
    }

    {
    const DollarCent d(-99,12);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-98.88", "print $-98.88");
    }

    {
    const DollarCent d(3,20);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$3.20", "print $3.20");
    }

    {
    const DollarCent d(-3,-20);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-3.20", "print $-3.20");
    }

    {
    const DollarCent d(3,-20);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$2.80", "print $2.80");
    }

    {
    const DollarCent d(-3,20);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$-2.80", "print $-2.80");
    }

    {
    const DollarCent d(10000000);
    std::ostringstream ostr;
    ostr << d;
    t.test(ostr.str() == "$10000000.00", "print $10000000.00");
    }

    {
    const DollarCent d(1,2);
    std::ostringstream ostr;
    ostr << d << d;
    t.test(ostr.str() == "$1.02$1.02", "print $1.02 twice");
    }
}


// test_DollarCent_equality
// Test suite for class DollarCent: equality
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent_equality(Tester & t)
{
    std::cout << "Test Suite: class DollarCent - equality"
              << std::endl;

    // operator==

    {
    const DollarCent d1(1,20);
    const DollarCent d2(1,20);
    bool result = (d1 == d2);
    t.test(result, "op== - equality");
    }

    {
    const DollarCent d1(1,20);
    const DollarCent d2(1,22);
    bool result = (d1 == d2);
    t.test(!result, "op== - cents inequality");
    }

    {
    const DollarCent d1(1,20);
    const DollarCent d2(2,20);
    bool result = (d1 == d2);
    t.test(!result, "op== - dollars inequality");
    }

    {
    const DollarCent d1(-1,-20);
    const DollarCent d2(-1,-20);
    bool result = (d1 == d2);
    t.test(result, "op== - equality of negative");
    }
    {
    const DollarCent d1(-1,-20);
    const DollarCent d2(-1,-22);
    bool result = (d1 == d2);
    t.test(!result, "op== - cents inequality of negative");
    }

    {
    const DollarCent d1(-1,-20);
    const DollarCent d2(-2,-20);
    bool result = (d1 == d2);
    t.test(!result, "op== - dollars inequality of negative");
    }

    // operator!=

    {
    const DollarCent d1(1,20);
    const DollarCent d2(1,20);
    bool result = (d1 != d2);
    t.test(!result, "op!= - equality");
    }

    {
    const DollarCent d1(1,20);
    const DollarCent d2(1,22);
    bool result = (d1 != d2);
    t.test(result, "op!= - cents inequality");
    }

    {
    const DollarCent d1(1,20);
    const DollarCent d2(2,20);
    bool result = (d1 != d2);
    t.test(result, "op!= - dollars inequality");
    }

    {
    const DollarCent d1(-1,-20);
    const DollarCent d2(-1,-20);
    bool result = (d1 != d2);
    t.test(!result, "op!= - equality of negative");
    }
    {
    const DollarCent d1(-1,-20);
    const DollarCent d2(-1,-22);
    bool result = (d1 != d2);
    t.test(result, "op!= - cents inequality of negative");
    }

    {
    const DollarCent d1(-1,-20);
    const DollarCent d2(-2,-20);
    bool result = (d1 != d2);
    t.test(result, "op!= - dollars inequality of negative");
    }

}


// test_DollarCent_ordering
// Test suite for class DollarCent: ordering
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent_ordering(Tester & t)
{
    std::cout << "Test Suite: class DollarCent - ordering"
              << std::endl;

    // operator<

    {
    DollarCent d1(1,20);
    DollarCent d2(2,5);
    bool result = (d1 < d2);
    t.test(result, "op< - positive < positive - #1");
    }

    {
    DollarCent d1(1,20);
    DollarCent d2(1,21);
    bool result = (d1 < d2);
    t.test(result, "op< - positive < positive - #2");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(1,5);
    bool result = (d1 < d2);
    t.test(!result, "op< - positive > positive");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(2,20);
    bool result = (d1 < d2);
    t.test(!result, "op< - positive == positive");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-2,-5);
    bool result = (d1 < d2);
    t.test(!result, "op< - negative > negative - #1");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-1,-21);
    bool result = (d1 < d2);
    t.test(!result, "op< - negative > negative - #2");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-1,-5);
    bool result = (d1 < d2);
    t.test(result, "op< - negative < negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-2,-20);
    bool result = (d1 < d2);
    t.test(!result, "op< - negative == negative");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(-2,-20);
    bool result = (d1 < d2);
    t.test(!result, "op< - positive, negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(2,20);
    bool result = (d1 < d2);
    t.test(result, "op< - negative, positive");
    }

    // operator<=

    {
    DollarCent d1(1,20);
    DollarCent d2(2,5);
    bool result = (d1 <= d2);
    t.test(result, "op<= - positive < positive - #1");
    }

    {
    DollarCent d1(1,20);
    DollarCent d2(1,21);
    bool result = (d1 <= d2);
    t.test(result, "op<= - positive < positive - #2");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(1,5);
    bool result = (d1 <= d2);
    t.test(!result, "op<= - positive > positive");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(2,20);
    bool result = (d1 <= d2);
    t.test(result, "op<= - positive == positive");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-2,-5);
    bool result = (d1 <= d2);
    t.test(!result, "op<= - negative > negative - #1");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-1,-21);
    bool result = (d1 <= d2);
    t.test(!result, "op<= - negative > negative - #2");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-1,-5);
    bool result = (d1 <= d2);
    t.test(result, "op<= - negative < negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-2,-20);
    bool result = (d1 <= d2);
    t.test(result, "op<= - negative == negative");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(-2,-20);
    bool result = (d1 <= d2);
    t.test(!result, "op<= - positive, negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(2,20);
    bool result = (d1 <= d2);
    t.test(result, "op<= - negative, positive");
    }

    // operator>

    {
    DollarCent d1(1,20);
    DollarCent d2(2,5);
    bool result = (d1 > d2);
    t.test(!result, "op> - positive < positive - #1");
    }

    {
    DollarCent d1(1,20);
    DollarCent d2(1,21);
    bool result = (d1 > d2);
    t.test(!result, "op> - positive < positive - #2");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(1,5);
    bool result = (d1 > d2);
    t.test(result, "op> - positive > positive");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(2,20);
    bool result = (d1 > d2);
    t.test(!result, "op> - positive == positive");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-2,-5);
    bool result = (d1 > d2);
    t.test(result, "op> - negative > negative - #1");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-1,-21);
    bool result = (d1 > d2);
    t.test(result, "op> - negative > negative - #2");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-1,-5);
    bool result = (d1 > d2);
    t.test(!result, "op> - negative < negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-2,-20);
    bool result = (d1 > d2);
    t.test(!result, "op> - negative == negative");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(-2,-20);
    bool result = (d1 > d2);
    t.test(result, "op> - positive, negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(2,20);
    bool result = (d1 > d2);
    t.test(!result, "op> - negative, positive");
    }

    // operator>=

    {
    DollarCent d1(1,20);
    DollarCent d2(2,5);
    bool result = (d1 >= d2);
    t.test(!result, "op>= - positive < positive - #1");
    }

    {
    DollarCent d1(1,20);
    DollarCent d2(1,21);
    bool result = (d1 >= d2);
    t.test(!result, "op>= - positive < positive - #2");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(1,5);
    bool result = (d1 >= d2);
    t.test(result, "op>= - positive > positive");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(2,20);
    bool result = (d1 >= d2);
    t.test(result, "op>= - positive == positive");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-2,-5);
    bool result = (d1 >= d2);
    t.test(result, "op>= - negative > negative - #1");
    }

    {
    DollarCent d1(-1,-20);
    DollarCent d2(-1,-21);
    bool result = (d1 >= d2);
    t.test(result, "op>= - negative > negative - #2");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-1,-5);
    bool result = (d1 >= d2);
    t.test(!result, "op>= - negative < negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(-2,-20);
    bool result = (d1 >= d2);
    t.test(result, "op>= - negative == negative");
    }

    {
    DollarCent d1(2,20);
    DollarCent d2(-2,-20);
    bool result = (d1 >= d2);
    t.test(result, "op>= - positive, negative");
    }

    {
    DollarCent d1(-2,-20);
    DollarCent d2(2,20);
    bool result = (d1 >= d2);
    t.test(!result, "op>= - negative, positive");
    }
}


// test_DollarCent_addsub
// Test suite for class DollarCent: addition & subtraction
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent_addsub(Tester & t)
{
    std::cout << "Test Suite: class DollarCent - addition & subtraction"
              << std::endl;

    {
    DollarCent d1(1, 2);
    const DollarCent d2(3, 4);
    const DollarCent d3(4, 6);
    const DollarCent & d4(d1 += d2);

    t.test(d1 == d3, "op+= - pos+=pos - value of 1st operand");
    t.test(d4 == d3, "op+= - pos+=pos - return value");

    const DollarCent d5(7, 10);
    d1 += d2;
    t.test(d4 == d5, "op+= - pos+=pos - return is reference");
    }

    {
    DollarCent d1(1, 2);
    const DollarCent d2(-3, -4);
    const DollarCent d3(-2, -2);
    const DollarCent & d4(d1 += d2);

    t.test(d1 == d3, "op+= - pos+=neg - value of 1st operand");
    t.test(d4 == d3, "op+= - pos+=neg - return value");

    const DollarCent d5(-5, -6);
    d1 += d2;
    t.test(d4 == d5, "op+= - pos+=neg - return is reference");
    }

    {
    DollarCent d1(1, 2);
    const DollarCent d2(3, 4);
    const DollarCent d3(-2, -2);
    const DollarCent & d4(d1 -= d2);

    t.test(d1 == d3, "op-= - pos-=pos - value of 1st operand");
    t.test(d4 == d3, "op-= - pos-=pos - return value");

    const DollarCent d5(-5, -6);
    d1 -= d2;
    t.test(d4 == d5, "op-= - pos-=pos - return is reference");
    }

    {
    DollarCent d1(1, 2);
    const DollarCent d2(-3, -4);
    const DollarCent d3(4, 6);
    const DollarCent & d4(d1 -= d2);

    t.test(d1 == d3, "op-= - pos-=neg - value of 1st operand");
    t.test(d4 == d3, "op-= - pos-=neg - return value");

    const DollarCent d5(7, 10);
    d1 -= d2;
    t.test(d4 == d5, "op-= - pos-=neg - return is reference");
    }

    {
    const DollarCent d1(1, 2);
    const DollarCent d2(3, 4);
    const DollarCent d3(4, 6);
    const DollarCent d4 = d1 + d2;

    t.test(d4 == d3, "op+ - pos+pos - return value");
    }

    {
    const DollarCent d1(1, 2);
    const DollarCent d2(-3, -4);
    const DollarCent d3(-2, -2);
    const DollarCent d4 = d1 + d2;

    t.test(d4 == d3, "op+ - pos+neg - return value");
    }

    {
    const DollarCent d1(1, 2);
    const DollarCent d2(3, 4);
    const DollarCent d3(-2, -2);
    const DollarCent d4 = d1 - d2;

    t.test(d4 == d3, "op- - pos-pos - return value");
    }

    {
    const DollarCent d1(1, 2);
    const DollarCent d2(-3, -4);
    const DollarCent d3(4, 6);
    const DollarCent d4 = d1 - d2;

    t.test(d4 == d3, "op- - pos-neg - return value");
    }
}


// test_DollarCent_unaryMinus
// Test suite for class DollarCent: unary minus
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent_unaryMinus(Tester & t)
{
    std::cout << "Test Suite: class DollarCent - unary minus"
              << std::endl;

    {
    DollarCent d1(1, 2);
    DollarCent d2(-1, -2);
    t.test(-d1 == d2, "unary op- - positive value");
    }

    {
    DollarCent d1(-1, -2);
    DollarCent d2(1, 2);
    t.test(-d1 == d2, "unary op- - negative value");
    }

    {
    DollarCent d1;
    t.test(-d1 == d1, "unary op- - zero value");
    }
}


// test_DollarCent_nonRef
// Test suite for class DollarCent: verifying non-reference returns
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent_nonRef(Tester & t)
{
    std::cout << "Test Suite: class DollarCent -"
              << " verifying non-reference returns"
              << std::endl;

    // Actually these tests should not even compile if any of the
    //  relevant functions return by reference.

    {
    DollarCent a, b;
    const DollarCent x(1, 2);
    decltype(a+b) y(x);
    y.setDC(0, 0);
    t.test(x != y, "binary op+ - return is not by reference");
    }

    {
    DollarCent a, b;
    const DollarCent x(1, 2);
    decltype(a-b) y(x);
    y.setDC(0, 0);
    t.test(x != y, "binary op- - return is not by reference");
    }

    {
    DollarCent a;
    const DollarCent x(1, 2);
    decltype(-a) y(x);
    y.setDC(0, 0);
    t.test(x != y, "unary op- - return is not by reference");
    }

    {
    DollarCent a(1, 2);
    DollarCent b;
    const bool x = true;
    decltype(a==b) y(x);
    y = false;
    t.test(x != y, "op== - return is not by reference");
    } 

    {
    DollarCent a(1, 2);
    DollarCent b;
    const bool x = true;
    decltype(a!=b) y(x);
    y = false;
    t.test(x != y, "op!= - return is not by reference");
    } 

    {
    DollarCent a(1, 2);
    DollarCent b;
    const bool x = true;
    decltype(a<b) y(x);
    y = false;
    t.test(x != y, "op< - return is not by reference");
    } 

    {
    DollarCent a(1, 2);
    DollarCent b;
    const bool x = true;
    decltype(a<=b) y(x);
    y = false;
    t.test(x != y, "op<= - return is not by reference");
    } 

    {
    DollarCent a(1, 2);
    DollarCent b;
    const bool x = true;
    decltype(a>b) y(x);
    y = false;
    t.test(x != y, "op> - return is not by reference");
    } 

    {
    DollarCent a(1, 2);
    DollarCent b;
    const bool x = true;
    decltype(a>=b) y(x);
    y = false;
    t.test(x != y, "op>= - return is not by reference");
    } 
}


// test_DollarCent
// Test suite for class DollarCent
// Uses other test-suite functions
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// Does not throw (No-Throw Guarantee)
void test_DollarCent(Tester & t)
{
    // Do all the test suites
    std::cout << "TEST SUITES FOR CLASS DollarCent" << std::endl;
    test_DollarCent_output(t);
    test_DollarCent_equality(t);
    test_DollarCent_ordering(t);
    test_DollarCent_addsub(t);
    test_DollarCent_unaryMinus(t);
    test_DollarCent_nonRef(t);
}


// ************************************************************************
// Main program
// ************************************************************************


// main
// Runs class DollarCent test suite, prints results to cout.
int main()
{
    Tester t;
    test_DollarCent(t);

    std::cout << std::endl;
    if (t.allPassed())
    {
        std::cout << "All tests successful" 
                  << std::endl;
    }
    else
    {
        std::cout << "Tests ********** UNSUCCESSFUL **********"
                  << std::endl;
    }
    std::cout << std::endl;

    std::cout << "Press ENTER to quit ";
    while (std::cin.get() != '\n') ;

    return 0;
}

